@@ -26,67 +26,40 @@
 #include "session.h"
 #include "cache.h"
 
-#define MAX_CACHE_INDEX 27 * 27
+#define MAX_CACHE_INDEX 250
+
+extern char *hs_conlen;
 
 static t_cached_object *cache[MAX_CACHE_INDEX];
 static pthread_mutex_t cache_mutex[MAX_CACHE_INDEX];
-static pthread_mutex_t cachesize_mutex;
-static volatile off_t cachesize;
+static pthread_mutex_t cache_size_mutex;
+static volatile off_t cache_size;
 
 void init_cache_module(void) {
-	int i;
+	int index;
 
-	for (i = 0; i < MAX_CACHE_INDEX; i++) {
-		cache[i] = NULL;
-		pthread_mutex_init(&cache_mutex[i], NULL);
+	for (index = 0; index < MAX_CACHE_INDEX; index++) {
+		cache[index] = NULL;
+		pthread_mutex_init(&cache_mutex[index], NULL);
 	}
 
-	pthread_mutex_init(&cachesize_mutex, NULL);
-	cachesize = 0;
+	pthread_mutex_init(&cache_size_mutex, NULL);
+	cache_size = 0;
 }
 
 static int cache_index(char *filename) {
-	char *c, c1, c2;
+	int index = 0;
 
 	if (filename == NULL) {
 		return -1;
 	}
 
-	c = filename + strlen(filename);
-	while ((*c != '/') && (c > filename)) {
-		c--;
-	}
-	if (*c == '/') {
-		c++;
+	while (*filename != '\0') {
+		index += (unsigned char)*filename;
+		filename++;
 	}
 
-	if (*c == '\0') {
-		return -1;
-	} else {
-		c1 = *c;
-		if ((c1 >= 'A') && (c1 <= 'Z')) {
-			c1 -= ('A' - 1);
-		} else if ((c1 >= 'a') && (c1 <= 'z')) {
-			c1 -= ('a' - 1);
-		} else {
-			c1 = 0;
-		}
-	}
-	
-	if (*(c + 1) == '\0') {
-		c2 = 0;
-	} else {
-		c2 = *(c + 1);
-		if ((c2 >= 'A') && (c2 <= 'Z')) {
-			c2 -= ('A' - 1);
-		} else if ((c2 >= 'a') && (c2 <= 'z')) {
-			c2 -= ('a' - 1);
-		} else {
-			c2 = 0;
-		}
-	}
-
-	return (c1 * 27) + c2;
+	return abs(index) % MAX_CACHE_INDEX;
 }
 
 static t_cached_object *remove_from_cache(t_cached_object *object, int index) {
@@ -99,9 +72,9 @@ static t_cached_object *remove_from_cache(t_cached_object *object, int index) {
 		object->next->prev = object->prev;
 	}
 
-	pthread_mutex_lock(&cachesize_mutex);
-	cachesize -= object->size;
-	pthread_mutex_unlock(&cachesize_mutex);
+	pthread_mutex_lock(&cache_size_mutex);
+	cache_size -= object->size;
+	pthread_mutex_unlock(&cache_size_mutex);
 
 	if (object == cache[index]) {
 		cache[index] = object->next;
@@ -114,11 +87,38 @@ static t_cached_object *remove_from_cache(t_cached_object *object, int index) {
 	return next;
 }
 
-t_cached_object *add_to_cache(t_session *session, char *file) {
+static bool add_object_to_cache(t_cached_object *object) {	
+	int index;
+
+	if (object == NULL) {
+		return false;
+	} else if ((index = cache_index(object->file)) == -1) {
+		return false;
+	}
+
+	pthread_mutex_lock(&cache_mutex[index]);
+
+	object->prev = NULL;
+	object->next = cache[index];
+	if (cache[index] != NULL) {
+		cache[index]->prev = object;
+	}
+	cache[index] = object;
+
+	pthread_mutex_lock(&cache_size_mutex);
+	cache_size += object->size;
+	pthread_mutex_unlock(&cache_size_mutex);
+
+	pthread_mutex_unlock(&cache_mutex[index]);
+
+	return true;
+}
+
+t_cached_object *add_file_to_cache(t_session *session, char *file) {
 	t_cached_object *object;
 	struct stat status;
 	off_t size;
-	int fd, i;
+	int fd;
 	size_t bytes_read, bytes_total = 0;
 
 	if (file == NULL) {
@@ -129,7 +129,7 @@ t_cached_object *add_to_cache(t_session *session, char *file) {
 		return NULL;
 	} else if ((size < session->config->cache_min_filesize) || (size > session->config->cache_max_filesize)) {
 		return NULL;
-	} else if (cachesize + size > session->config->cache_size) {
+	} else if (cache_size + size > session->config->cache_size) {
 		return NULL;
 	} else if ((object = (t_cached_object*)malloc(sizeof(t_cached_object))) == NULL) {
 		return NULL;
@@ -170,9 +170,10 @@ t_cached_object *add_to_cache(t_session *session, char *file) {
 	object->deadline = session->time + TIME_IN_CACHE;
 	object->size = size;
 	object->in_use = 1;
+	object->type = cot_file;
 	copy_ip(&(object->last_ip), &(session->ip_address));
 
-	if ((i = cache_index(file)) == -1) {
+	if (add_object_to_cache(object) == false) {
 		free(object->data);
 		free(object->file);
 		free(object);
@@ -180,45 +181,28 @@ t_cached_object *add_to_cache(t_session *session, char *file) {
 		return NULL;
 	}
 
-	pthread_mutex_lock(&cache_mutex[i]);
-
-	object->prev = NULL;
-	object->next = cache[i];
-	if (cache[i] != NULL) {
-		cache[i]->prev = object;
-	}
-	cache[i] = object;
-
-	pthread_mutex_lock(&cachesize_mutex);
-	cachesize += object->size;
-	pthread_mutex_unlock(&cachesize_mutex);
-
-	pthread_mutex_unlock(&cache_mutex[i]);
-
 	return object;
 }
 
-t_cached_object *search_cache(t_session *session, char *file) {
+t_cached_object *search_cache_for_file(t_session *session, char *file) {
 	t_cached_object *object, *result = NULL;
 	struct stat status;
 	off_t size;
-	int i;
+	int index;
 
 	if (file == NULL) {
 		return NULL;
 	} else if ((size = filesize(file)) == -1) {
 		return NULL;
-	}
-
-	if ((i = cache_index(file)) == -1) {
+	} else if ((index = cache_index(file)) == -1) {
 		return NULL;
 	}
 
-	pthread_mutex_lock(&cache_mutex[i]);
+	pthread_mutex_lock(&cache_mutex[index]);
 
-	object = cache[i];
+	object = cache[index];
 	while (object != NULL) {
-		if (object->size == size) {
+		if ((object->type == cot_file) && (object->size == size)) {
 			if (strcmp(object->file, file) == 0) {
 				if (stat(file, &status) == 0) {
 					if ((object->deadline > session->time) && (status.st_mtime == object->last_changed)) {
@@ -231,10 +215,10 @@ t_cached_object *search_cache(t_session *session, char *file) {
 						object->in_use++;
 						result = object;
 					} else if (object->in_use <= 0) {
-						remove_from_cache(object, i);
+						remove_from_cache(object, index);
 					}
 				} else if (object->in_use <= 0) {
-					remove_from_cache(object, i);
+					remove_from_cache(object, index);
 				}
 				break;
 			}
@@ -242,11 +226,160 @@ t_cached_object *search_cache(t_session *session, char *file) {
 		object = object->next;
 	}
 
-	pthread_mutex_unlock(&cache_mutex[i]);
+	pthread_mutex_unlock(&cache_mutex[index]);
 
 	return result;
 }
 
+static char *make_url(t_session *session, char *request_uri) {
+	char *url;
+	size_t len;
+
+	len = strlen(*(session->host->hostname.item));
+	if (request_uri == NULL) {
+		request_uri = session->request_uri;
+	}
+
+	if ((url = (char*)malloc(len + strlen(request_uri) + 1)) == NULL) {
+		return NULL;
+	}
+
+	memcpy(url, *(session->host->hostname.item), len);
+	strcpy(url + len, request_uri);
+
+	return url;
+}
+
+t_cached_object *add_cgi_output_to_cache(t_session *session, char *cgi_output, int size, int time) {
+	t_cached_object *object;
+	size_t len;
+	char *pos, *data, str[50];
+
+	if ((pos = strstr(cgi_output, "\r\n\r\n")) == NULL) {
+		return NULL;
+	}
+	*pos = '\0';
+	if (strcasestr(cgi_output, hs_conlen) == NULL) {
+		/* Output has no Content-Length
+		 */
+		len = size - (pos + 4 - cgi_output);
+		sprintf(str, "%s%ld\r\n", hs_conlen, len);
+		len = strlen(str);
+		if ((data = (char*)malloc(len + size)) == NULL) {
+			*pos = '\r';
+			return NULL;
+		}
+		memcpy(data, str, len);
+		memcpy(data + len, cgi_output, size);
+		size += len;
+	} else {
+		/* Output has Content-Length
+		 */
+		if ((data = (char*)malloc(size)) == NULL) {
+			*pos = '\r';
+			return NULL;
+		}
+		memcpy(data, cgi_output, size);
+	}
+	*pos = '\r';
+
+	if ((object = (t_cached_object*)malloc(sizeof(t_cached_object))) == NULL) {
+		free(data);
+		return NULL;
+	} else if ((object->file = make_url(session, NULL)) == NULL) {
+		free(object);
+		free(data);
+		return NULL;
+	}
+
+	object->deadline = session->time + time;
+	object->size = size;
+	object->in_use = 0;
+	object->type = cot_cgi;
+	object->data = data;
+	copy_ip(&(object->last_ip), &(session->ip_address));
+
+	if (add_object_to_cache(object) == false) {
+		free(object->data);
+		free(object->file);
+		free(object);
+
+		return NULL;
+	}
+
+	return object;
+}
+
+t_cached_object *search_cache_for_cgi_output(t_session *session) {
+	t_cached_object *object, *result = NULL;
+	char *url;
+	int index;
+
+	if ((url = make_url(session, NULL)) == NULL) {
+		return NULL;
+	} else if ((index = cache_index(url)) == -1) {
+		free(url);
+		return NULL;
+	}
+
+	pthread_mutex_lock(&cache_mutex[index]);
+
+	object = cache[index];
+	while (object != NULL) {
+		if (object->type == cot_cgi) {
+			if (strcmp(object->file, url) == 0) {
+				if (object->deadline > session->time) {
+					object->in_use++;
+					result = object;
+				} else if (object->in_use <= 0) {
+					remove_from_cache(object, index);
+				}
+				break;
+			}
+		}
+		object = object->next;
+	}
+
+	pthread_mutex_unlock(&cache_mutex[index]);
+
+	free(url);
+
+	return result;
+}
+
+void remove_cgi_output_from_cache(t_session *session, char *request_uri) {
+	t_cached_object *object;
+	char *url;
+	int index;
+
+	if ((url = make_url(session, request_uri)) == NULL) {
+		return;
+	} else if ((index = cache_index(url)) == -1) {
+		free(url);
+		return;
+	}
+
+	pthread_mutex_lock(&cache_mutex[index]);
+
+	object = cache[index];
+	while (object != NULL) {
+		if (object->type == cot_cgi) {
+			if (strcmp(object->file, url) == 0) {
+				if (object->in_use <= 0) {
+					remove_from_cache(object, index);
+				} else {
+					object->deadline = 0;
+				}
+				break;
+			}
+		}
+	}
+
+	pthread_mutex_unlock(&cache_mutex[index]);
+
+	free(url);
+}
+
 void done_with_cached_object(t_cached_object *object, bool remove_object) {
 	if (remove_object) {
 		object->deadline = 0;
@@ -256,16 +389,16 @@ void done_with_cached_object(t_cached_object *object, bool remove_object) {
 
 void check_cache(time_t now) {
 	t_cached_object *object;
-	int i;
+	int index;
 
-	for (i = 0; i < MAX_CACHE_INDEX; i++) {
-		pthread_mutex_lock(&cache_mutex[i]);
+	for (index = 0; index < MAX_CACHE_INDEX; index++) {
+		pthread_mutex_lock(&cache_mutex[index]);
 
-		object = cache[i];
+		object = cache[index];
 		while (object != NULL) {
 			if (now > object->deadline) {
 				if (object->in_use <= 0) {
-					object = remove_from_cache(object, i);
+					object = remove_from_cache(object, index);
 					continue;
 				} else {
 					object->deadline = 0;
@@ -274,11 +407,11 @@ void check_cache(time_t now) {
 			object = object->next;
 		}
 
-		pthread_mutex_unlock(&cache_mutex[i]);
+		pthread_mutex_unlock(&cache_mutex[index]);
 	}
 }
 
-int clear_cache(void) {
+int clear_cache() {
 	t_cached_object *object, *list;
 	int i, removed = 0;
 
@@ -294,9 +427,9 @@ int clear_cache(void) {
 			if (object->in_use == 0) {
 				/* Object unused, so remove
 				 */
-				pthread_mutex_lock(&cachesize_mutex);
-				cachesize -= object->size;
-				pthread_mutex_unlock(&cachesize_mutex);
+				pthread_mutex_lock(&cache_size_mutex);
+				cache_size -= object->size;
+				pthread_mutex_unlock(&cache_size_mutex);
 				
 				free(object->data);
 				free(object->file);
@@ -326,15 +459,17 @@ void print_cache_list(FILE *fp) {
 
 	now = time(NULL);
 
+	/* File cache
+	 */
 	for (i = 0; i < MAX_CACHE_INDEX; i++) {
 		pthread_mutex_lock(&cache_mutex[i]);
 
 		object = cache[i];
 		while (object != NULL) {
-			fprintf(fp, "  Filename : %s\n", object->file);
-			fprintf(fp, "  File size: %lld bytes\n", (long long)object->size);
+			fprintf(fp, "  Filename  : %s\n", object->file);
+			fprintf(fp, "  File size : %lld bytes\n", (long long)object->size);
 			if ((secs = object->deadline - now) > 0) {
-				fprintf(fp, "  Time left: %d seconds\n\n", secs);
+				fprintf(fp, "  Time left : %d seconds\n\n", secs);
 			} else {
 				fprintf(fp, "  File marked for removal from cache\n\n");
 			}
@@ -346,11 +481,11 @@ void print_cache_list(FILE *fp) {
 		pthread_mutex_unlock(&cache_mutex[i]);
 	}
 
-	fprintf(fp, "  Total: %lld bytes in %d files.\n", (long long)size, files);
+	fprintf(fp, "  Total: %lld bytes in %d cache objects.\n", (long long)size, files);
 }
 
 off_t size_of_cache(void) {
-	return cachesize;
+	return cache_size;
 }
 #endif
 
