@@ -25,10 +25,10 @@
 
 void init_client_module(void);
 int  add_client(t_session *session);
+int  reposition_client(t_session *session, t_ip_addr *ip_address);
 int  mark_client_for_removal(t_session *session, int delay);
 void check_remove_deadlines(t_config *config, time_t now);
 int  remove_client(t_session *session, bool free_session);
-void check_write_blocks_on_socket(int max_write_block_time);
 int  connection_allowed(t_ip_addr *ip, int maxperip, int maxtotal);
 int  disconnect_clients(t_config *config);
 bool client_is_flooding(t_session *session);
@@ -38,7 +38,7 @@ int  kick_client(int id);
 int  kick_ip(t_ip_addr *ip);
 
 int  ban_ip(t_ip_addr *ip, int timer, bool kick_on_ban);
-int  reban_ip(t_ip_addr *ip);
+void reban_ip(t_ip_addr *ip);
 void check_ban_list(t_config *config, time_t now);
 int  unban_ip(t_ip_addr *ip);
 
@@ -49,9 +49,9 @@ void close_client_sockets_for_cgi_run(void);
 
 #ifdef HAVE_COMMAND
 void print_client_list(FILE *fp);
-int number_of_clients(void);
+int  number_of_clients(void);
 void print_ban_list(FILE *fp);
-int number_of_bans(void);
+int  number_of_bans(void);
 #endif
 
 #endif
