@@ -26,6 +26,6 @@
 
 #define TASK_RUNNER_INTERVAL 10
 
-#define TIMER_OFF -1
+#define TIMER_OFF   0
 
 #endif
