@@ -227,8 +227,12 @@ static bool basic_http_authentication(t_session *session, char *auth_str) {
 					} else {
 						register_wrong_password(session);
 					}
+				} else {
+					register_wrong_password(session);
 				}
 				free(passwd);
+			} else  {
+				register_wrong_password(session);
 			}
 		}
 	}
@@ -322,6 +326,7 @@ static bool digest_http_authentication(t_session *session, char *auth_str) {
 	/* Retrieve A1 from passwordfile
 	 */
 	if ((passwd = get_A1(session, username, realm)) == NULL) {
+		register_wrong_password(session);
 		return false;
 	} else if (strlen(passwd) != 32) {
 		free(passwd);
@@ -333,6 +338,7 @@ static bool digest_http_authentication(t_session *session, char *auth_str) {
 	/* Group oke?
 	 */
 	if (group_oke(session, username, &(session->host->required_group)) == false) {
+		register_wrong_password(session);
 		return false;
 	}
 	
