@@ -298,6 +298,10 @@ int ssl_send(SSL *ssl, char *buffer, unsigned int length) {
 	return -1;
 }
 
+int ssl_pending(SSL *ssl) {
+	return SSL_pending(ssl);
+}
+
 int ssl_close(SSL **ssl) {
 	int result;
 
