@@ -27,7 +27,7 @@ void log_garbage(t_session *session);
 void log_exploit_attempt(t_session *session, char *type, char *data);
 void log_unban(char *logfile, t_ip_addr *ip_address, unsigned long connect_attempts);
 void log_cgi_error(t_session *session, char *mesg);
-void close_logfiles(t_host *host, bool force_close);
+void close_logfiles(t_host *host, time_t now);
 void close_logfiles_for_cgi_run(t_host *host);
 #ifdef DEBUG
 void log_debug(t_session *session, char *mesg, ...);
