@@ -36,7 +36,7 @@ static const struct {
 	const char *text;
 } sqli_detection[] = {
 	{"union\\s*select"},
-	{"[\\s'0-9a-z]\\s*--\\s"},
+	{"[\\s'0-9a-z]\\s*--\\s*.*"},
 	{"'\\s*(and|or|&&|\\|\\|)\\s*('|[0-9]|[a-z.]*\\s*=)"},
 	{NULL}
 };
