@@ -397,8 +397,8 @@ int send_file(t_session *session) {
 					cached_object = NULL;
 				} else
 #endif
-					if ((cached_object = search_cache(session, session->file_on_disk)) == NULL) {
-						cached_object = add_to_cache(session, session->file_on_disk);
+					if ((cached_object = search_cache_for_file(session, session->file_on_disk)) == NULL) {
+						cached_object = add_file_to_cache(session, session->file_on_disk);
 					}
 
 				if (cached_object != NULL) {
@@ -454,10 +454,16 @@ int send_file(t_session *session) {
  */
 int execute_cgi(t_session *session) {
 	int retval = 200, result, handle, len;
-	char *line, *new_line, *contyp_start, *contyp_end, *code, c;
+	char *line, *new_line, *str_begin, *str_end, *code, c;
 	bool in_body = false, send_in_chunks = true, wrap_cgi;
 #ifdef CYGWIN
 	char *old_path, *win32_path;
+#endif
+#ifdef HAVE_CACHE
+	t_cached_object *cached_object;
+	char *cache_buffer = NULL;
+	int  cache_size = 0;
+	int  cache_time = 0;
 #endif
 	t_cgi_result cgi_result;
 	t_connect_to *connect_to;
@@ -601,6 +607,24 @@ int execute_cgi(t_session *session) {
 		}
 	}
 
+#ifdef HAVE_CACHE
+	/* Search for CGI output in cache
+	 */
+	if (session->request_method == GET) {
+		if ((cached_object = search_cache_for_cgi_output(session)) != NULL) {
+			if (send_header(session) == -1) {
+				retval = rs_DISCONNECT;
+			} else if (send_buffer(session, cached_object->data, cached_object->size) == -1) {
+				retval = rs_DISCONNECT;
+			}
+
+			done_with_cached_object(cached_object, false);
+
+			return retval;
+		}
+	}
+#endif
+
 	cgi_info.type = session->cgi_type;
 	cgi_info.input_buffer_size = cgi_info.error_buffer_size = CGI_BUFFER_SIZE;
 	cgi_info.input_len = cgi_info.error_len = 0;
@@ -719,6 +743,21 @@ int execute_cgi(t_session *session) {
 								retval = rs_DISCONNECT;
 							}
 						}
+
+#ifdef HAVE_CACHE
+						/* Add body content to cache buffer
+						 */
+						if ((cache_buffer != NULL) && (retval == 200)) {
+							if (cache_size + cgi_info.input_len > session->config->cache_max_filesize) {
+								free(cache_buffer);
+								cache_buffer = NULL;
+							} else {
+								memcpy(cache_buffer	+ cache_size, cgi_info.input_buffer, cgi_info.input_len);
+								cache_size += cgi_info.input_len;
+								*(cache_buffer + cache_size) = '\0';
+							}
+						}
+#endif
 						cgi_info.input_len = 0;
 					} else {
 						/* Read HTTP header
@@ -739,13 +778,13 @@ int execute_cgi(t_session *session) {
 						if (new_line != NULL) {
 							*(new_line + 2) = '\0';
 							if (session->throttle == 0) {
-								if ((contyp_start = strcasestr(cgi_info.input_buffer, hs_contyp)) != NULL) {
-									if ((contyp_end = strchr(contyp_start, '\r')) != NULL) {
-										contyp_start += 14;
-										c = *contyp_end;
-										*contyp_end = '\0';
-										session->throttle = get_throttlespeed(contyp_start, session->config->throttle);
-										*contyp_end = c;
+								if ((str_begin = strcasestr(cgi_info.input_buffer, hs_contyp)) != NULL) {
+									if ((str_end = strchr(str_begin, '\r')) != NULL) {
+										str_begin += 14;
+										c = *str_end;
+										*str_end = '\0';
+										session->throttle = get_throttlespeed(str_begin, session->config->throttle);
+										*str_end = c;
 									}
 								}
 							}
@@ -756,6 +795,60 @@ int execute_cgi(t_session *session) {
 								}
 							}
 
+#ifdef HAVE_CACHE
+							/* Look for store-in-cache CGI header
+							 */
+							if (session->request_method == GET) {
+								if ((str_begin = strcasestr(cgi_info.input_buffer, "\nX-Hiawatha-Cache:")) != NULL) {
+									str_begin += 18;
+									while (*str_begin == ' ') {
+										str_begin++;
+									}
+
+									str_end = str_begin;
+									while ((*str_end != '\r') && (*str_end != '\0')) {
+										str_end++;
+									}
+									if (*str_end == '\r') {
+										*str_end = '\0';
+										cache_time = str2int(str_begin);
+										*str_end = '\r';
+
+										if (cache_time != -1) {
+											if (cache_time < MIN_CGI_CACHE_TIMER) {
+												cache_time = MIN_CGI_CACHE_TIMER;
+											} else if (cache_time > MAX_CGI_CACHE_TIMER) {
+												cache_time = MAX_CGI_CACHE_TIMER;
+											}
+
+											if ((cache_buffer = (char*)malloc(session->config->cache_max_filesize + 1)) != NULL) {
+												*(cache_buffer + session->config->cache_max_filesize) = '\0';
+											}
+										}
+									}
+								}
+							}
+
+							/* Look for remove-from-cache CGI header
+							 */
+							if ((str_begin = strcasestr(cgi_info.input_buffer, "\nX-Hiawatha-Cache-Remove:")) != NULL) {
+								str_begin += 25;
+								while (*str_begin == ' ') {
+									str_begin++;
+								}
+
+								str_end = str_begin;
+								while ((*str_end != '\r') && (*str_end != '\0')) {
+									str_end++;
+								}
+								if (*str_end == '\r') {
+									*str_end = '\0';
+									remove_cgi_output_from_cache(session, str_begin);
+									*str_end = '\r';
+								}
+							}
+#endif
+
 							if (strcasestr(cgi_info.input_buffer, "\nLocation:") != NULL) {
 								session->return_code = 302;
 								if (send_header(session) == -1) {
@@ -826,10 +919,11 @@ int execute_cgi(t_session *session) {
 							}
 							session->header_sent = true;
 
+							len = cgi_info.input_len - len;
 							/* Send first part of the body
 							 */
 							if (session->request_method != HEAD) {
-								if ((len = cgi_info.input_len - len) > 0) {
+								if (len > 0) {
 									if (send_in_chunks) {
 										result = send_chunk(session, new_line, len);
 									} else {
@@ -848,6 +942,24 @@ int execute_cgi(t_session *session) {
 */
 							}
 
+#ifdef HAVE_CACHE
+							/* Add header to cache buffer
+							 */
+							if (cache_buffer != NULL) {
+								if (retval != 200) {
+									free(cache_buffer);
+									cache_buffer = NULL;
+								} else if (cache_size + cgi_info.input_len > session->config->cache_max_filesize) {
+									free(cache_buffer);
+									cache_buffer = NULL;
+								} else {
+									memcpy(cache_buffer + cache_size, cgi_info.input_buffer, cgi_info.input_len);
+									cache_size += cgi_info.input_len;
+									*(cache_buffer + cache_size) = '\0';
+								}
+							}
+#endif
+
 							in_body = true;
 							cgi_info.input_len = 0;
 						} else if (cgi_info.input_len > MAX_CGI_HEADER) {
@@ -877,6 +989,17 @@ int execute_cgi(t_session *session) {
 		} /* switch */
 	} while (retval == 200);
 
+#ifdef HAVE_CACHE
+	/* Add cache buffer to cache
+	 */
+	if (cache_buffer != NULL) {
+		if (retval == rs_QUIT) {
+			add_cgi_output_to_cache(session, cache_buffer, cache_size, cache_time);
+		}
+		free(cache_buffer);
+	}
+#endif
+
 	if (session->cgi_type == fastcgi) {
 		close(cgi_info.from_cgi);
 	} else {
