@@ -1,7 +1,14 @@
-/* Hiawatha | userconfig.c
+/* This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; version 2 of the License. For a copy,
+ * see http://www.gnu.org/licenses/gpl-2.0.html.
  *
- * All the routines for loading the configuration files
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+ * GNU General Public License for more details.
  */
+
 #include "config.h"
 #include <stdio.h>
 #include <stdlib.h>
@@ -19,6 +26,10 @@ void init_groups(t_groups *groups) {
 int parse_userid(char *userid, uid_t *uid) {
 	struct passwd *pwd;
 
+	if (userid == NULL) {
+		return -1;
+	}
+
 	if ((*uid = str2int(userid)) == -1) {
 		if ((pwd = getpwnam(userid)) == NULL) {
 			return -1;
@@ -35,6 +46,10 @@ int parse_userid(char *userid, uid_t *uid) {
 static gid_t parse_groupid(char *groupid, gid_t *gid) {
 	struct passwd *pwd;
 
+	if (groupid == NULL) {
+		return -1;
+	}
+
 	if ((*gid = str2int(groupid)) == -1) {
 		if ((pwd = getpwnam(groupid)) == NULL) {
 			return -1;
@@ -53,6 +68,10 @@ int parse_groups(char *groupid, gid_t *gid, t_groups *groups) {
 	gid_t *id;
 	char *c;
 
+	if ((groupid == NULL) || (gid == NULL) || (groups == NULL)) {
+		return -1;
+	}
+
 	c = groupid;
 	while (*c != '\0') {
 		if (*c == ',') {
@@ -87,6 +106,10 @@ int parse_groups(char *groupid, gid_t *gid, t_groups *groups) {
 }
 
 static bool is_member(char *user, char **group) {
+	if ((user == NULL) || (group == NULL)) {
+		return false;
+	}
+
 	while (*group != NULL) {
 		if (strcmp(user, *group) == 0) {
 			return true;
@@ -103,9 +126,12 @@ int lookup_group_ids(uid_t uid, gid_t *gid, t_groups *groups) {
 	int size, result;
 	gid_t *id;
 
-	if ((pwd = getpwuid(uid)) == NULL) {
+	if ((gid == NULL) || (groups == NULL)) {
+		return -1;
+	} else if ((pwd = getpwuid(uid)) == NULL) {
 		return -1;
 	}
+
 	if ((*gid = pwd->pw_gid) == 0) {
 		result = 0;
 	} else {
